/**
 * A helper that checks if the given argument is of a certain type.
 */
export function $is(type: BigIntConstructor, arg: any): arg is bigint;
export function $is(type: BooleanConstructor, arg: any): arg is boolean;
export function $is(type: NumberConstructor, arg: any): arg is number;
export function $is(type: StringConstructor, arg: any): arg is string;
export function $is(type: SymbolConstructor, arg: any): arg is symbol;
export function $is(type: undefined, arg: any): arg is null | undefined;
export function $is<T extends abstract new (...args: any[]) => any>(type: T, arg: any): arg is InstanceType<T>;

/**
 * A replacement for the `in` operator (not to be confused with the `for-in` loop) that works properly.
 */
export function $in(key: any, ref: any): boolean;

/**
 * A helper that checks equality of the given arguments.
 */
export function $eq(x: any, y: any): boolean;

export type $HArgs = [
  keyof HTMLElementTagNameMap | Element | ShadowRoot,
  ...(Record<string, any> | Element | ShadowRoot | Text | string | number | null | undefined | $HArgs)[]
];

/**
 * A lightweight [HyperScript](https://github.com/hyperhype/hyperscript)-style helper for creating and modifying `HTMLElement`s.
 *
 * Arguments of type `null` or `undefined` are simply ignored.
 *
 * Arguments of type `Record<string, any>` are mappings of attributes and properties.
 * Keys starting with `$` specify *properties* to be set on the element being created or modified (without the leading `$`).
 * All other keys specify *attributes* to be set by `setAttribute`.
 * An attribute equal to `false` causes the attribute to be removed by `removeAttribute`.
 *
 * All other args are nodes to be appended to the element being created or modified.
 */
export function $h<T extends keyof HTMLElementTagNameMap>(tag: T, ...args: $HArgs[1][]): HTMLElementTagNameMap[T];
export function $h<E extends Element>(element: E, ...args: $HArgs[1][]): E;
export function $h<S extends ShadowRoot>(shadowRoot: S, ...args: $HArgs[1][]): S;
export function $h(tagOrElementOrShadowRoot: $HArgs[0], ...args: $HArgs[1][]): Element | ShadowRoot;

export type $SArgs = [
  keyof SVGElementTagNameMap | SVGElement,
  ...(Record<string, any> | SVGElement | Text | string | number | null | undefined | $SArgs)[]
];

/**
 * A lightweight [HyperScript](https://github.com/hyperhype/hyperscript)-style helper for creating and modifying `SVGElement`s (cf. `$h`).
 */
export function $s<T extends keyof SVGElementTagNameMap>(tag: T, ...args: $SArgs[1][]): SVGElementTagNameMap[T];
export function $s<E extends SVGElement>(svgElement: E, ...args: $SArgs[1][]): E;
export function $s(tagOrSvgElement: $SArgs[0], ...args: $SArgs[1][]): SVGElement;

export type $CNode = {
  [attributeOrSelector: string]: string | number | $CNode;
};

/**
 * A simple CSS-in-JS helper.
 *
 * The `root` parameter provides a hierarchical description of CSS rules.
 *
 * - Keys of sub-objects whose values are NOT objects are treated as CSS attribute, and values are treated as values of those CSS attributes; the concatenation of keys of all parent objects is a CSS rule.
 * - All keys ignore the part starting with a splitter (default: `$$`) sign until the end of the key (e.g. `src$$1` → `src`, `@font-face$$1` → `@font-face`).
 * - In keys specifying CSS attribute, all uppercase letters are replaced by lowercase letters with an additional `-` character preceding them (e.g. `fontFamily` → `font-family`).
 * - Commas in keys that makes a CSS rule cause it to “split” and create separate rules for each part (e.g. `{div:{margin:1,'.a,.b,.c':{margin:2}}}` → `div{margin:1}div.a,div.b,div.c{margin:2}`).
 * - Top-level keys that begin with `@` are not concatenated with sub-object keys.
 */
export function $c(root: Record<string, $CNode>, splitter?: string): string;

/**
 * A generic helper for syntax highlighting.
 */
export function $highlight(pattern: RegExp, highlighters: ((chunk: string, index: number) => $HArgs[2])[], innerText: string): $HArgs[2][];

/**
 * Language translations helper.
 */
export function $locale(
  locales: Record<string, Record<string, string | Record<string, string>>>,
  defaultLanguage: string,
  languages?: readonly string[]
): (text?: string | undefined, version?: string | undefined) => string;
