/**
 * @template {abstract new (...args: any[]) => any} T
 *
 * @type {{
 *   (type: BigIntConstructor,  arg: any): arg is bigint;
 *   (type: BooleanConstructor, arg: any): arg is boolean;
 *   (type: NumberConstructor,  arg: any): arg is number;
 *   (type: StringConstructor,  arg: any): arg is string;
 *   (type: SymbolConstructor,  arg: any): arg is symbol;
 *   (type: undefined,          arg: any): arg is undefined | null;
 *   <T extends abstract new (...args: any[]) => any>(type: T, arg: any): arg is InstanceType<T>;
 * }}
 *
 * @returns {arg is bigint | boolean | number | string | symbol | undefined | null | InstanceType<T>}
 */
export const $is = (/** @type {T} */ type, /** @type {any} */ arg) => arg?.constructor === type

export const $in = (/** @type {any} */ key, /** @type {any} */ ref) =>
  ($is(String, key) || $is(Number, key) || $is(Symbol, key)) && Object.hasOwnProperty.call(ref ?? Object, key)

export const $eq = /** @returns {boolean} */ (/** @type {any} */ x, /** @type {any} */ y) => {
  if (x === y) {
    return true
  }

  const xConstructor = x?.constructor

  if (xConstructor === y?.constructor) {
    if (xConstructor === Number) {
      return (isNaN(x) && isNaN(y)) || +x === +y
    }

    if (xConstructor === Date) {
      return +x === +y
    }

    if (xConstructor === String || xConstructor === RegExp) {
      return '' + x === '' + y
    }

    if (xConstructor === Array) {
      return x.length === y.length && x.every((/** @type {any} */ item, /** @type {number} */ index) => $eq(item, y[index]))
    }

    if (xConstructor === Object) {
      const keysOfX = Object.keys(x)
      return keysOfX.length === Object.keys(y).length && keysOfX.every(key => $eq(x[key], y[key]))
    }

    if (xConstructor === Set || xConstructor === Map) {
      if (x.size !== y.size) {
        return false
      }
      const xa = [...x]
      const ya = [...y]
      return xa.every(xv => ya.find(yv => $eq(xv, yv)))
    }
  }

  return false
}

/**
 * @typedef {[
 *   keyof HTMLElementTagNameMap| Element | ShadowRoot,
 *   ...(Record<string, any> | Element | ShadowRoot | Text | string | number | null | undefined | $HArgs)[]
 * ]} $HArgs
 */

/**
 * @type {{
 *   <T extends keyof HTMLElementTagNameMap>(tag: T,                              ...args: $HArgs[1][]): HTMLElementTagNameMap[T];
 *   <E extends Element>                    (element: E,                          ...args: $HArgs[1][]): E;
 *   <S extends ShadowRoot>                 (shadowRoot: S,                       ...args: $HArgs[1][]): S;
 *                                          (tagOrElementOrShadowRoot: $HArgs[0], ...args: $HArgs[1][]): Element | ShadowRoot;
 * }}
 *
 * @returns {Element | ShadowRoot}
 */
export const $h = (/** @type {$HArgs[0]} */ tagOrElementOrShadowRoot, /** @type {$HArgs[1][]} */ ...args) => {
  const elementOrShadowRoot = $is(String, tagOrElementOrShadowRoot) ? document.createElement(tagOrElementOrShadowRoot) : tagOrElementOrShadowRoot

  for (const arg of args) {
    let node = null

    if (arg instanceof Element || arg instanceof ShadowRoot || arg instanceof Text) {
      node = arg
    } else if ($is(String, arg) || $is(Number, arg)) {
      // @ts-expect-error:
      node = document.createTextNode(arg)
    } else if ($is(Array, arg)) {
      // @ts-expect-error:
      node = $h(...arg)
    } else if (arg != null) {
      for (const name in arg) {
        const value = arg[name]
        if (name[0] === '$') {
          const name1 = name.slice(1)
          if (name1 in elementOrShadowRoot && $is(Object, value)) { // warning: $in cannot be used here!
            for (const name2 in value) {
              // @ts-expect-error:
              elementOrShadowRoot[name1][name2] = value[name2]
            }
          } else {
            // @ts-expect-error:
            elementOrShadowRoot[name1] = value
          }
        } else if (elementOrShadowRoot instanceof Element) {
          if (value === true) {
            elementOrShadowRoot.setAttribute(name, '')
          } else if (value === false) {
            elementOrShadowRoot.removeAttribute(name)
          } else {
            elementOrShadowRoot.setAttribute(name, value)
          }
        }
      }
    }

    if (node != null) {
      elementOrShadowRoot.appendChild(node)
    }
  }

  return elementOrShadowRoot
}

const /** @type {Record<string, string>} */ NS = {
  xlink: 'http://www.w3.org/1999/xlink'
}

/**
 * @typedef {[
 *   keyof SVGElementTagNameMap | SVGElement,
 *   ...(Record<string, any> | SVGElement | Text | string | number | null | undefined | $SArgs)[]
 * ]} $SArgs
 */

/**
 * @type {{
 *   <T extends keyof SVGElementTagNameMap>(tag: T,                     ...args: $SArgs[1][]): SVGElementTagNameMap[T];
 *   <E extends SVGElement>                (svgElement: E,              ...args: $SArgs[1][]): E;
 *                                         (tagOrSvgElement: $SArgs[0], ...args: $SArgs[1][]): SVGElement;
 * }}
 *
 * @returns {SVGElement}
 */
export const $s = (/** @type {$SArgs[0]} */ tagOrSvgElement, /** @type {$SArgs[1][]} */ ...args) => {
  const svgElement = $is(String, tagOrSvgElement) ? document.createElementNS('http://www.w3.org/2000/svg', tagOrSvgElement) : tagOrSvgElement

  for (const arg of args) {
    let node = null

    if (arg instanceof SVGElement || arg instanceof Text) {
      node = arg
    } else if ($is(String, arg) || $is(Number, arg)) {
      // @ts-expect-error:
      node = document.createTextNode(arg)
    } else if ($is(Array, arg)) {
      // @ts-expect-error:
      node = $s(...arg)
    } else if (arg != null) {
      for (const name in arg) {
        const value = arg[name]

        if (name[0] === '$') {
          const name1 = name.slice(1)
          if (name1 in svgElement && $is(Object, value)) { // warning: $in cannot be used here!
            for (const name2 in value) {
              // @ts-expect-error:
              svgElement[name1][name2] = value[name2]
            }
          } else {
            // @ts-expect-error:
            svgElement[name1] = value
          }
        } else {
          const indexOfColon = name.indexOf(':')

          if (indexOfColon >= 0) {
            const /** @type {string=} */ ns = NS[name.slice(0, indexOfColon)]

            if (ns != null) {
              const basename = name.slice(indexOfColon + 1)

              if (value === true) {
                svgElement.setAttributeNS(ns, basename, '')
              } else if (value === false) {
                svgElement.removeAttributeNS(ns, basename)
              } else {
                svgElement.setAttributeNS(ns, basename, $is(String, value) ? value : '' + value)
              }
            }
          } else {
            if (value === true) {
              svgElement.setAttribute(name, '')
            } else if (value === false) {
              svgElement.removeAttribute(name)
            } else {
              svgElement.setAttribute(name, $is(String, value) ? value : '' + value)
            }
          }
        }
      }
    }

    if (node != null) {
      svgElement.appendChild(node)
    }
  }

  return svgElement
}

/**
 * @typedef {{
 *   [attributeOrSelector: string]: string | number | $CNode;
 * }} $CNode
 */

const _c = (
  /** @type {$CNode}   */ node,
  /** @type {string}   */ prefix,
  /** @type {string[]} */ result,
  /** @type {(text: string) => string} */ split
) => {
  const /** @type {[$CNode | string[], string][]} */ queue = [[node, prefix]]

  while (queue.length) {
    const [style2, prefix2] = queue.shift() ?? []

    if (style2 == null || prefix2 == null) {
      continue
    }

    if ($is(Array, style2)) {
      result.push(prefix2, prefix2 !== '' ? '{' : '', style2.join(';'), prefix2 !== '' ? '}' : '')
    } else {
      const /** @type {[$CNode | string[], string][]} */ todo = []
      let /** @type {string[]} */ attributes = []
      let attributesPushed = false

      for (const key in style2) {
        const value = style2[key]

        if ($is(String, value) || $is(Number, value)) {
          if (!attributesPushed) {
            attributesPushed = true
            attributes = []
            todo.push([attributes, prefix2])
          }

          attributes.push(`${split(key).replace(/([A-Z])/g, (_, letter) => '-' + letter.toLowerCase())}:${value}`)
        } else {
          attributesPushed = false

          const /** @type {string[]} */ newPrefix = []
          const keySplitted = key.split(',')

          for (const prefixItem of prefix2.split(',')) {
            for (const keyItem of keySplitted) {
              newPrefix.push(prefixItem + keyItem)
            }
          }

          todo.push([value, newPrefix.join(',')])
        }
      }

      queue.unshift(...todo)
    }
  }
}

export const $c = (/** @type {Record<string, $CNode>} */ root, /** @type {string} */ splitter = '$$') => {
  const split = (/** @type {string} */ text) => text.split(splitter)[0]
  const /** @type {string[]} */ result = []

  for (const key in root) {
    const value = root[key]

    if (key[0] === '@') {
      result.push(split(key) + '{')
      _c(value, '', result, split)
      result.push('}')
    } else {
      _c(value, split(key), result, split)
    }
  }

  return result.join('')
}

export const $highlight = (
  /** @type {RegExp} */ pattern,
  /** @type {((chunk: string, index: number) => $HArgs[1])[]} */ highlighters,
  /** @type {string} */ innerText
) => {
  const /** @type {$HArgs[1][]} */ result = []

  innerText.split(pattern).forEach((chunk, index) => {
    index %= highlighters.length
    if (chunk != null) {
      result.push(highlighters[index](chunk, index))
    }
  })

  return result
}

const _locale = (
  /** @type {Record<string, Record<string, string | Record<string, string>>>} */ locales,
  /** @type {string} */ language,
  /** @type {string} */ text,
  /** @type {string} */ version
) => {
  // @ts-expect-error:
  const v = locales?.[language]?.[version]?.[text]

  if ($is(String, v)) {
    return v
  }

  const t = locales?.[language]?.[text]

  return $is(String, t) ? t : text
}

export const $locale = (
  /** @type {Record<string, Record<string, string | Record<string, string>>>} */ locales,
  /** @type {string} */ defaultLanguage,
  /** @type {readonly string[]} */ languages = navigator.languages
) => {
  for (const language of languages) {
    if ($in(language, locales)) {
      // @ts-expect-error:
      return _locale.bind(0, locales, language)
    }
  }

  // @ts-expect-error:
  return _locale.bind(0, locales, defaultLanguage)
}
